import re

import polars as pl
import pytest
from polars.testing import assert_frame_equal
from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from src.sparkpolars.config import Config
from src.sparkpolars.sparkpolars import ModeMethod, to_spark, toPolars


def test_spark_to_polars_native(spark_session, spark_df, polars_df):
    assert_frame_equal(toPolars(spark_df, lazy=False), polars_df)
    assert_frame_equal(toPolars(spark_df, lazy=True), polars_df.lazy())

    data_spark = to_spark(polars_df, config=Config(map_elements=["cin"]))
    assert data_spark.schema == spark_df.schema
    assert data_spark.collect() == spark_df.collect()

    data_spark = to_spark(polars_df.lazy(), config=Config(map_elements=["cin"]))
    assert data_spark.schema == spark_df.schema
    assert data_spark.collect() == spark_df.collect()

    with pytest.raises(NotImplementedError, match=re.escape("Method not implemented.")):
        toPolars(spark_df, mode="not implemented")
    with pytest.raises(NotImplementedError, match=re.escape("Method not implemented.")):
        to_spark(polars_df, mode="not implemented")


def test_spark_to_polars_using_arrow(spark_session, spark_df, polars_df):
    pl_data = toPolars(
        spark_df,
        lazy=False,
        mode=ModeMethod.ARROW,
    )
    assert_frame_equal(pl_data, polars_df)


def test_spark_to_polars_using_pandas(spark_session, spark_df, polars_df):
    pl_data = toPolars(
        spark_df.drop("cin", "address"),  # Struct and Map types are not supported in pandas
        lazy=False,
        mode=ModeMethod.PANDAS,
    )
    assert_frame_equal(pl_data, polars_df.drop("cin", "address"))


def test_polars_to_spark_using_pandas(spark_session, spark_df, polars_df):
    data_spark = to_spark(
        polars_df.drop("cin"),  # pandas does not support Map type
        mode=ModeMethod.PANDAS,
    )
    assert data_spark.schema == spark_df.drop("cin").schema
    assert data_spark.collect() == spark_df.drop("cin").collect()


def test_polars_to_spark_using_arrow(spark_session, spark_df, polars_df):
    data_spark = to_spark(
        polars_df.drop("cin"),
        mode=ModeMethod.ARROW,
    )
    assert data_spark.schema == spark_df.drop("cin").schema
    assert data_spark.collect() == spark_df.drop("cin").collect()


def test_empty_spark_dataframe(spark_session):
    schema = StructType([StructField("a", IntegerType()), StructField("b", StringType())])
    df = spark_session.createDataFrame([], schema=schema)
    pl_df = df.toPolars()

    expected_polars = pl.DataFrame(
        schema={
            "a": pl.Int32,
            "b": pl.Utf8,
        },
        data={
            "a": [],
            "b": [],
        },
    )

    assert_frame_equal(pl_df, expected_polars)


def test_empty_polars_dataframe(spark_session):

    pl_df = pl.DataFrame(
        schema={
            "a": pl.Int32,
            "b": pl.Utf8,
        },
        data={
            "a": [],
            "b": [],
        },
    )
    df = pl_df.to_spark()

    schema = StructType([StructField("a", IntegerType()), StructField("b", StringType())])
    expected_df = spark_session.createDataFrame([], schema=schema)

    assert df.schema == expected_df.schema
    assert df.collect() == expected_df.collect()



import pytest
import os
from pyspark.sql import SparkSession, Row
import pyspark.sql.types as T
import polars as pl
from datetime import datetime, UTC
from zoneinfo import ZoneInfo

@pytest.fixture(scope="session")
def spark_session():
    os.environ["TZ"] = "UTC"
    spark = SparkSession.builder \
        .appName("pytest") \
        .config("spark.sql.timezone", "UTC") \
        .config("spark.sql.session.timeZone", "UTC") \
        .getOrCreate()

    yield spark
    spark.stop()
    
@pytest.fixture()
def spark_data():
    data =  [
       Row(name='Alice', age=20, is_student=True, gpa=3.5, courses=['Math', 'Physics'], address=Row(street='123 Main St', city='Springfield', state='IL', zip=62701), created_at=datetime(2020, 1, 1, 1, 1, 1, tzinfo=UTC), updated_at=datetime(2020, 1, 1, 1, 1, 1, tzinfo=UTC), hobby=[['Skiing', 'Hiking'], ['Reading', 'Writing']], cin={"hello": "world", "another": "fine", "this": "is"}),
       Row(name='Bob', age=21, is_student=False, gpa=3.0, courses=['Math', 'Physics'], address=Row(street='123 Main St', city='Springfield', state='IL', zip=62701), created_at=datetime(2020, 1, 1, 1, 1, 1, tzinfo=UTC), updated_at=datetime(2020, 1, 1, 1, 1, 1, tzinfo=UTC), hobby=[['Skiing', 'Hiking'], ['Reading', 'Writing']], cin={"hello": "world", "another": "fine"}),
    ]
    
    return data
    
@pytest.fixture()
def polars_data():
    data = [
       {
           'name': 'Alice',
           'age': 20,
           'is_student': True,
           'gpa': 3.5,
           'courses': ['Math','Physics'],
           'address': {'street': '123 Main St', 'city': 'Springfield','state': 'IL','zip': 62701},
           'created_at': datetime(2020, 1, 1, 1, 1, 1, tzinfo=UTC),
           'updated_at': datetime(2020, 1, 1, 1, 1, 1, tzinfo=UTC),
           'hobby': [['Skiing', 'Hiking'], ['Reading', 'Writing']],
           "cin": [{'key': 'this', 'value': 'is'}, {'key': 'hello', 'value': 'world'}, {'key': 'another', 'value': 'fine'}],
       },
       {
           'name': 'Bob',
           'age': 21,
           'is_student': False,
           'gpa': 3.0,
           'courses': ['Math','Physics'],
           'address': {'street': '123 Main St', 'city': 'Springfield','state': 'IL','zip': 62701},
           'created_at': datetime(2020, 1, 1, 1, 1, 1, tzinfo=UTC),
           'updated_at': datetime(2020, 1, 1, 1, 1, 1, tzinfo=UTC),
           'hobby': [['Skiing', 'Hiking'], ['Reading', 'Writing']],
           "cin": [{'key': 'another', 'value': 'fine'}, {'key': 'hello', 'value': 'world'}],
       },
    ]
    
    return data
@pytest.fixture()
def schema_spark():
    schema=T.StructType(
        [
            T.StructField("name", T.StringType()),
            T.StructField("age", T.IntegerType()),
            T.StructField("is_student", T.BooleanType()),
            T.StructField("gpa", T.FloatType()),
            T.StructField("courses", T.ArrayType(T.StringType())),
            T.StructField("address", T.StructType([
                T.StructField("street", T.StringType()),
                T.StructField("city", T.StringType()),
                T.StructField("state", T.StringType()),
                T.StructField("zip", T.LongType())
            ])),
            T.StructField("created_at", T.TimestampType()),
            T.StructField("updated_at", T.TimestampType()),
            T.StructField("hobby", T.ArrayType(T.ArrayType(T.StringType()))),
            T.StructField("cin", T.MapType(T.StringType(), T.StringType()))
        ]
    )
    return schema
    
@pytest.fixture()
def spark_df(spark_session, schema_spark, spark_data):
    df = spark_session.createDataFrame(
        schema=schema_spark,
        data = spark_data
    )
    
    return df

@pytest.fixture()
def schema_polars():
    schema={
        'name': pl.String,
        'age': pl.Int32,
        'is_student': pl.Boolean,
        'gpa': pl.Float32,
        'courses': pl.List(pl.String),
        'address': pl.Struct({'street': pl.String, 'city': pl.String, 'state': pl.String, 'zip': pl.Int64}),
        'created_at': pl.Datetime(time_unit='us', time_zone='UTC'),
        'updated_at': pl.Datetime(time_unit='us', time_zone='UTC'),
        'hobby': pl.List(pl.List(pl.String)),
        "cin": pl.List(pl.Struct({'key': pl.String, 'value': pl.String})),
    }
    return schema

@pytest.fixture()
def polars_df(schema_polars, polars_data):
    df = pl.DataFrame(
        schema=schema_polars,
        data = polars_data
    )
    
    return df


from typing import Any

from polars.datatypes import (
    Binary,
    Boolean,
    Date,
    Datetime,
    Decimal,
    Float32,
    Float64,
    Int8,
    Int16,
    Int32,
    Int64,
    List,
    String,
    Struct,
)
from polars.datatypes import (
    DataType as PolarsDataTypes,
)
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import Row as SparkRow
from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    ByteType,
    DateType,
    DecimalType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    MapType,
    ShortType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from pyspark.sql.types import (
    DataType as SparkDataTypes,
)

from .config import Config

SIMPLE_TYPES: dict = {
    StringType(): String,
    ByteType(): Int8,
    ShortType(): Int16,
    IntegerType(): Int32,
    LongType(): Int64,
    BooleanType(): Boolean,
    FloatType(): Float32,
    DateType(): Date,
    DoubleType(): Float64,
    BinaryType(): Binary,
}


def _type_convert_pyspark_to_polars(
    pyspark_type: SparkDataTypes,
    config: Config | None = None,
    tz: str | None = None,
) -> PolarsDataTypes:
    """
    Recursively converts PySpark types to Polars types.

    :param pyspark_type: The PySpark type to convert

    :param config: The configuration of the application. Default is None.

    :return: The equivalent Polars type
    """
    if isinstance(pyspark_type, ArrayType):
        return List(_type_convert_pyspark_to_polars(pyspark_type.elementType, config, tz))
    if isinstance(pyspark_type, StructType):
        return Struct(
            {
                field.name: _type_convert_pyspark_to_polars(field.dataType, config, tz)
                for field in pyspark_type.fields
            },
        )
    if isinstance(pyspark_type, MapType):
        return List(
            Struct(
                {
                    "key": _type_convert_pyspark_to_polars(pyspark_type.keyType),
                    "value": _type_convert_pyspark_to_polars(pyspark_type.valueType),
                },
            ),
        )
    if isinstance(pyspark_type, TimestampType):
        if config is None:
            return Datetime(time_zone=tz)
        return Datetime(time_unit=config.time_unit, time_zone=tz)
    if isinstance(pyspark_type, DecimalType):
        return Decimal(pyspark_type.precision, pyspark_type.scale)
    polar_type = SIMPLE_TYPES.get(pyspark_type)
    if polar_type:
        return polar_type
    msg = f"Unsupported type: {pyspark_type!r}"
    raise ValueError(msg)


def _convert_schema_spark_to_polars(
    df_schema_field: list[StructField],
    config: Config | None = None,
    tz: str | None = None,
) -> dict[str, Any]:
    """
    Converts a PySpark schema to a Polars schema.

    :param df_schema_field: list of Structfield

    :param config: The configuration of the application. Default is None.

    :return: The Polars schema
    """
    return {
        field.name: _type_convert_pyspark_to_polars(field.dataType, config, tz)
        for field in df_schema_field
    }


def _unpack_map(packed_dict: dict) -> list[dict]:
    """
    This method converts a dictionary into a list of {'key': ..., 'value': ...} dictionaries.

    :param packed_dict: The dictionary to convert

    :return: The list of dictionaries
    """
    return [{"key": k, "value": v} for k, v in packed_dict.items()]


def _spark_row_as_dict(row: SparkRow, config: Config | None = None) -> dict[str, Any]:
    """
    Converts a Spark Row to a dictionary.

    :param row: The Spark Row

    :param config: The configuration of the application. Default is None.

    :return: The dictionary representation of the Spark Row
    """
    if not isinstance(row, SparkRow):
        msg = "Expected a Spark Row"
        raise TypeError(msg)
    return {
        key: (
            _spark_row_as_dict(value, config)
            if isinstance(value, SparkRow)
            else _unpack_map(value) if isinstance(value, dict) else value
        )
        for key, value in row.asDict().items()
    }


def _get_time_zone(spark_dataframe: SparkDataFrame) -> str:  # pragma: no cover
    """
    This function gets the time zone from the Spark DataFrame.

    :param spark_dataframe: The Spark DataFrame

    :return: The time zone
    """
    session_tz = spark_dataframe.sparkSession.conf.get("spark.sql.session.timeZone")
    if session_tz:
        return session_tz
    sql_tz = spark_dataframe.sparkSession.conf.get("spark.sql.timezone")
    if sql_tz:
        return sql_tz
    return "UTC"



from src.sparkpolars._from_polars_to_spark import _pack_map
from src.sparkpolars._from_spark_to_polars import _unpack_map


def test__pack_map():
    dict_list = [{"key": "a", "value": 1}, {"key": "b", "value": 2}]
    assert _pack_map(dict_list) == {"a": 1, "b": 2}


def test__unpack_map():
    dict_ = {"a": 1, "b": 2}
    assert _unpack_map(dict_) == [{"key": "a", "value": 1}, {"key": "b", "value": 2}]
    assert _unpack_map({}) == []



from importlib.metadata import version
from importlib.util import find_spec


def check_version_and_module(module_name: str, minimum_version: str) -> None:
    """
    This function checks if a module is installed and
        if its version is greater than or equal to the minimum version.

    :param module_name: The name of the module to check

    :param minimum_version: The minimum version required

    :return: None
    """
    spec = find_spec(module_name)
    if spec is None:
        msg = f"Module {module_name} is not installed."
        raise ImportError(msg)

    if version(module_name) <= minimum_version:
        msg = f"Module {module_name} version {minimum_version} or higher is required."
        raise ImportError(msg)



import re

import pytest

from src.sparkpolars._importlib_utils import check_version_and_module


def test_importlib_utils():

    check_version_and_module("pip", "1.0.0")
    with pytest.raises(ImportError, match=re.escape("Module package_not_existing is not installed")):
        check_version_and_module("package_not_existing", "1.0.0")
    with pytest.raises(ImportError, match=re.escape("Module pip version 99999.0.0 or higher is required")):
        check_version_and_module("pip", "99999.0.0")




import polars as pl
import pyspark.sql.types as T
import pytest

from src.sparkpolars._from_polars_to_spark import _type_convert_polars_to_spark
from src.sparkpolars._from_spark_to_polars import _type_convert_pyspark_to_polars
from src.sparkpolars.config import Config


@pytest.mark.parametrize(("pyspark_type", "expected_polars_type"), [
    (T.StringType(), pl.String),
    (T.IntegerType(), pl.Int32),
    (T.LongType(), pl.Int64),
    (T.ShortType(), pl.Int16),
    (T.ByteType(), pl.Int8),
    (T.BooleanType(), pl.Boolean),
    (T.BinaryType(), pl.Binary),
    (T.DecimalType(16, 10), pl.Decimal(16, 10)),
    (T.DecimalType(5, 5), pl.Decimal(5, 5)),
    (T.TimestampType(), pl.Datetime),
    (T.ArrayType(T.StringType()), pl.List(pl.String)),
    (T.ArrayType(T.ArrayType(T.StringType())), pl.List(pl.List(pl.String))),
    (T.MapType(T.StringType(), T.StringType()), pl.List(pl.Struct({"key": pl.String, "value": pl.String}))),
    (T.ArrayType(T.ArrayType(T.ArrayType(T.IntegerType()))), pl.List(pl.List(pl.List(pl.Int32)))),
    (T.StructType([T.StructField("street", T.StringType()), T.StructField("city", T.StringType()), T.StructField("state", T.StringType()), T.StructField("zip", T.LongType())]), pl.Struct({"street": pl.String, "city": pl.String, "state": pl.String, "zip": pl.Int64})),
    (T.StructType([T.StructField("street", T.StringType()), T.StructField("city", T.StringType()), T.StructField("state", T.ArrayType(T.ArrayType(T.StringType()))), T.StructField("zip", T.LongType())]), pl.Struct({"street": pl.String, "city": pl.String, "state": pl.List(pl.List(pl.String)), "zip": pl.Int64})),
])
def test__type_convert_pyspark_to_polars(pyspark_type, expected_polars_type):
    assert _type_convert_pyspark_to_polars(pyspark_type) == expected_polars_type


def test__type_convert_pyspark_to_polars_unknown():
    with pytest.raises(ValueError, match="Unsupported type: 'unknown'"):
        _type_convert_pyspark_to_polars("unknown")


def test__test_convert_spark_to_polars_datetime():
    assert _type_convert_pyspark_to_polars(T.TimestampType(), Config(time_unit="ns")) == pl.Datetime("ns")
    assert _type_convert_pyspark_to_polars(T.TimestampType(), tz="Africa/Tunis") == pl.Datetime(time_zone="Africa/Tunis")


@pytest.mark.parametrize(("is_map", "polar_type", "expected_spark_type"), [
    (False, pl.String, T.StringType()),
    (False, pl.Int32, T.IntegerType()),
    (False, pl.Int64, T.LongType()),
    (False, pl.Int16, T.ShortType()),
    (False, pl.Int8, T.ByteType()),
    (False, pl.Boolean, T.BooleanType()),
    (False, pl.Binary, T.BinaryType()),
    (False, pl.Datetime(), T.TimestampType()),
    (False, pl.Decimal(16, 10), T.DecimalType(16, 10)),
    (False, pl.Decimal(5, 5), T.DecimalType(5, 5)),
    (False, pl.List(pl.String), T.ArrayType(T.StringType())),
    (False, pl.List(pl.List(pl.String)), T.ArrayType(T.ArrayType(T.StringType()))),
    (False, pl.List(pl.List(pl.List(pl.Int32))), T.ArrayType(T.ArrayType(T.ArrayType(T.IntegerType())))),
    (None , pl.Struct({"street": pl.String, "city": pl.String, "state": pl.String, "zip": pl.Int64}), T.StructType([T.StructField("street", T.StringType()), T.StructField("city", T.StringType()), T.StructField("state", T.StringType()), T.StructField("zip", T.LongType())])),
    (False, pl.Struct({"street": pl.String, "city": pl.String, "state": pl.List(pl.List(pl.String)), "zip": pl.Int64}), T.StructType([T.StructField("street", T.StringType()), T.StructField("city", T.StringType()), T.StructField("state", T.ArrayType(T.ArrayType(T.StringType()))), T.StructField("zip", T.LongType())])),
    (True , pl.List(pl.Struct({"key": pl.String, "value": pl.String})), T.MapType(T.StringType(), T.StringType())),
])
def test__type_convert_polars_to_spark(is_map, polar_type, expected_spark_type):
    assert _type_convert_polars_to_spark(polar_type, is_map=is_map) == expected_spark_type


def test__type_convert_polars_to_spark_unknown():
    with pytest.raises(ValueError, match="Unsupported type: 'unknown'"):
        _type_convert_polars_to_spark("unknown")
    polar_type = pl.List(pl.Struct({"pey": pl.String, "value": pl.String}))
    with pytest.raises(ValueError, match=r"Only types List\(Struct\(key: Type, value: Type\)\) are supported for maps."):
        _type_convert_polars_to_spark(polar_type, is_map=True)
    polar_type = pl.List(pl.Struct({"key": pl.String, "falue": pl.String}))
    with pytest.raises(ValueError, match=r"Only types List\(Struct\(key: Type, value: Type\)\) are supported for maps."):
        _type_convert_polars_to_spark(polar_type, is_map=True)
    polar_type = pl.List(pl.Struct({"key": pl.String, "value": pl.String, "another": pl.String}))
    with pytest.raises(ValueError, match=r"Only types List\(Struct\(key: Type, value: Type\)\) are supported for maps."):
        _type_convert_polars_to_spark(polar_type, is_map=True)
    polar_type = pl.List(pl.List(pl.Int32))
    with pytest.raises(ValueError, match=r"Only types List\(Struct\(key: Type, value: Type\)\) are supported for maps."):
        _type_convert_polars_to_spark(polar_type, is_map=True)



"""
Setup file.
"""

import re
from pathlib import Path

from setuptools import find_packages, setup

with Path.open("src/sparkpolars/_version.py", encoding="utf-8") as version_file:
    version_groups = re.search(
        r'__version__ = "(\d+\.\d+\.\d+)"',
        version_file.read(),
        re.MULTILINE,
    )
    if version_groups:
        version = version_groups.group(1)
    else:
        msg = "Version not found in src/sparkpolars/_version.py"
        raise ValueError(msg)

setup(
    name="sparkpolars",
    version=version,
    packages=find_packages(exclude=["tests", "docs", "dist"]),
    install_requires=[
        "pyspark>=3.3.0",
        "polars>=1.00.0",
    ],
)
